#region "Class"
class Frontmatter {
    # regex for the yaml from a markdown file
    hidden [string]$_yamlRegex = '^(---\r?\n(?<yaml>.*?)\r?\n---\r?\n)?(?<content>.*)'
    # File associated with this frontmatter
    [System.IO.FileInfo]$FileInfo = $null
    hidden [string]$_originalFrontmatter = ""
    hidden [hashtable]$_yamlFrontmatter = @{}
    hidden [int]$_changeCount = 0

    Frontmatter([System.IO.FileInfo]$fileInfo) {
        $this.FileInfo = $fileInfo
        $this._originalFrontmatter = $this.GetFrontmatter()
        $this._yamlFrontmatter = $this.ParseFrontmatter()
    }

    [string]ToString() {
        return $this.FileInfo.Name
    }

    [string]GetFrontmatter() {

        $content = Get-Content -Path $this.FileInfo.FullName -Raw
        $match = [regex]::Match($content, $this._yamlRegex, [System.Text.RegularExpressions.RegexOptions]::Singleline)
        if ($match.Groups["yaml"].Success) {
            return $match.Groups["yaml"].Value
        }
        return ""
    }

    [hashtable]ParseFrontmatter() {
        if ($this._originalFrontmatter -eq "") {
            return @{}
        }
        return ConvertFrom-Yaml $this._originalFrontmatter
    }

    [int]ChangeCount() {
        return $this._changeCount
    }

    [bool]IsDirty() {
        return ($this._changeCount -gt 0)
    }

    [bool]HasFrontmatterChanged() {
        return $this._originalFrontmatter -ne $this.GetFrontmatter()
    }

    [void]BackupFile() {
        $backupFile = $this.FileInfo.DirectoryName + "\" + $this.FileInfo.BaseName + ".bak"
        # copy this file to the same name, but with .bak and forcefully overwrite any prior backup
        $this.FileInfo.CopyTo($backupFile, $true)
    }

    [int]ChangeFrontmatter([hashtable]$changeParameters,[scriptblock]$changeScript) {
        <#
        .SYNOPSIS
            Executes changes against the frontmatter yaml/hashtable and returns a count of the changes
        .DESCRIPTION
            By taking parameters in the form of a hashtable and the actual commands to execute in the form of a script it's possible for this class to handle many of the types of changes sent to it by different Cmdlets.
        .NOTES
            Script passed in usually takes the form of:

                # EXAMPLE: Remove a property from the frontmatter
                $cs = [scriptblock]{
                    param(
                        [hashtable]$yamlHt,
                        [hashtable]$changeParameters
                    )
                    # important to return a count greater than zero, to determine if the
                    # frontmatter has changed, and therefore when updated the file must be
                    # rewritten
                    [int]$c = 0
                    $changeParameters.Keys | ForEach-Object {
                        if ($yamlHt.ContainsKey($_)) {
                            $yamlHt.Remove($_)
                            $c++
                        }
                    }
                    return $c
                }

        #>

        $this._changeCount = Invoke-Command -ScriptBlock $changeScript -ArgumentList $this._yamlFrontmatter, $changeParameters
        return $this._changeCount

    }
    [void]UpdateMarkdown() {
        if ($this.IsDirty()) {

            # CBB: Could read the entire contents, process via regex and output from here
            # to save reading the file twice
            if ($this.HasFrontmatterChanged()) {
                Write-Warning -Message "Frontmatter has changed since the original file was read. Change aborted.`nFile: $($this.FileInfo.FullName)"
                return
            }

            # Ok to proceed here as the original yaml frontmatter hasn't changed
            [string]$newFrontmatter = ConvertTo-Yaml $this._yamlFrontmatter
            [string]$content = Get-Content -Path $this.FileInfo.FullName -Raw
            $match = [regex]::Match($content, $this._yamlRegex, [System.Text.RegularExpressions.RegexOptions]::Singleline)

            [string]$newContent = ""
            if ($newFrontmatter.Length -gt 0) {
                $newContent = "---`n$newFrontmatter`n---`n"
            }
            $newContent += $match.Groups["content"].Value

            Set-Content -Path $this.FileInfo.FullName -Value $newContent -Confirm:$false

            # when file is change, update IsDirty, not that it should be important
            $this._originalFrontmatter = $newFrontmatter
            $this._changeCount = 0
        }
    }
}
#endregion
#region "Test functions"
# function New-MarkdownFrontmatter([System.IO.FileInfo]$fileInfo) {
#     return [Frontmatter]::new($fileInfo)
# }
# Export-ModuleMember -Function New-MarkdownFrontmatter
#endregion
#region "Published functions"
function Get-MarkdownFrontmatter() {
    <#
    .SYNOPSIS
        Get the Yaml Frontmatter from markdown files
    .DESCRIPTION
        Gets the Yaml Frontmatter as the Frontmatter class. Can have individual properties returned from the Frontmatter if you wish.
    .NOTES
        The alias for this function is gmdf
    .EXAMPLE
        PS> Get-MarkdownFrontmatter -Path $VaultPath -Property "description","reviewstatus"

        description            Frontmatter                          reviewstatus
        -----------            -----------                          ------------
                            Original Note without frontmatter.md
        a bit more, here       This note has frontmatter more.md    Reviewed
        a small amount of text This note has frontmatter.md
        
        The example above gets the frontmatter from the files in the path and returns a Frontmatter class, as well as the properties requested
    #>
    
    [CmdletBinding()]
    param(
        # Specify the Path or List of Paths to process. Wildcards are supported.
        [Parameter(Mandatory=$true,ValueFromPipeline=$true,ValueFromPipelineByPropertyName=$true)]
        [string[]]$Path,
        # The properties to surface from the frontmatter Hashtable
        [string[]]$Property = $null,
        # prevent the default behaviour of recursing through subdirectories
        [switch]$NoRecurse
    )
    process {

        # Get the powershell-yaml module and ensure it's available
        If (!(Get-YamlModule)) {
            Write-Error -Message "Unable to get powershell-yaml module, exiting."
            return
        }

        foreach ($p in $Path) {
            Write-Verbose "Processing Path: $p with NoRecurse: $NoRecurse and Property: ""$Property"""
            $fi = Get-ChildItem -Path $p -Recurse:(!$NoRecurse) -File -Filter "*.md"
            foreach ($f in $fi) {
                $fm = [Frontmatter]::new($f)
                if ($null -eq $Property) {
                    $fm
                }
                else {

                    [hashtable]$output = @{}
                    $output["Frontmatter"] = $fm
                    foreach ($p in $Property) {
                        $output[$p] = $fm._yamlFrontmatter[$p]
                    }

                    $objOutput = [pscustomobject]$output
                    $objOutput
                }
            }
        }

        # return [Frontmatter]::new($Path)
    }
}
Set-Alias -Name gmdf -Value Get-MarkdownFrontmatter
Export-ModuleMember -Function Get-MarkdownFrontmatter -Alias gmdf

function Add-MarkdownFrontmatter() {
    <#
    .SYNOPSIS
        Adds properties to the frontmatter of markdown files
    .DESCRIPTION
        Adds properties to frontmatter if it doesn't exist, and assigns a value to it. If the property already exists, it will not be overwritten unless the Force switch is used. Can also be used to add items to a list in the frontmatter.
    .NOTES
        Use -Force to overwrite existing properties in the frontmatter. Without -Force the properties will only be created if they do not currently exist.

        This CmdLet supports -WhatIf and -Confirm to allow you to evaluate the changes before they are made.
    .EXAMPLE
        PS> Add-MarkdownFrontmatter -Path $VaultPath -Property "reviewstatus" -Value "To Be Reviewed"

        Will add the property "reviewstatus" to the frontmatter of all markdown files in the path, and assign the value "To Be Reviewed" to it. If the reviewstatus property already exists it will not be changed. Ideal for when you have started using a new property, but want to retrospectively add a "default" value that wasn't previously there.

    .EXAMPLE
        PS> Add-MarkdownFrontmatter -Path $VaultPath -Property "source" -Value "YouTube" -Force

        Will add (or overwrite) the property "source" on every markdown file in the folder structure 
    .EXAMPLE
        PS> Add-MarkdownFrontmatter -Path $VaultPath -Property "New List" -Value @() -Force

        Create or update the list "New List" in the frontmatter with an empty list of values

    .EXAMPLE
        PS> Add-MarkdownFrontmatter -Path $VaultPath -List "New List" -Item "YouTube","Book","Song","Website","Article"

        Add the items shown to the list "New List".
    #>
    
    
    [CmdletBinding(SupportsShouldProcess=$true)]
    param(
        # Specify the Path or List of Paths to process. Wildcards are supported.
        [Parameter(Mandatory=$true,ValueFromPipeline=$true,ValueFromPipelineByPropertyName=$true)]
        [string[]]$Path,
        # The property name to add to the frontmatter
        [Parameter(Mandatory=$true,ParameterSetName="Property")]
        [string]$Property,
        # The value to add (capability depends on the powershell-yaml module)
        [Parameter(Mandatory=$true,ParameterSetName="Property")]
        $Value,
        # When adding specific list items, specify the List and the Items to add
        [Parameter(Mandatory=$true,ParameterSetName="ListItem")]
        [string]$List,
        # When adding specific items to a list specify the list and the Items
        [Parameter(Mandatory=$true,ParameterSetName="ListItem")]
        [string[]]$Item,
        # prevent the default behaviour of recursing through subdirectories
        [switch]$NoRecurse,
        # switch to determine if backup should be skipped. By default every file that is modified will be copied into the same folder with the extension .bak before frontmatter is changed.
        [switch]$NoBackup,
        # Force the property (overwrite) even if it already exists
        [switch]$Force
    )

    process {
        [hashtable]$cp = @{}
        [scriptblock]$changeScript = $null
        [string]$updateDesc = ""

        Switch ($PSCmdLet.ParameterSetName)
        {
            "Property" {
                $updateDesc = "Add Property"
                $cp["Property"] = $Property
                $cp["Value"] = $Value
                $cp["Force"] = $Force
                $changeScript = [scriptblock]{
                    param(
                        [hashtable]$yamlHt,
                        [hashtable]$changeParameters
                    )
                    # important to return a count greater than zero, to determine if the
                    # frontmatter has changed, and therefore when updated the file must be
                    # rewritten
                    [int]$c = 0

                    if ($changeParameters["Force"]) {
                        $yamlHt[$changeParameters["Property"]] = $changeParameters["Value"]
                        $c++
                    }
                    else {
                        if (!($yamlHt.ContainsKey($changeParameters["Property"]))) {
                            $yamlHt[$changeParameters["Property"]] = $changeParameters["Value"]
                            $c++
                        }
                    }

                    return $c
                }
            }
            "ListItem" {
                $updateDesc = "Add List Item"
                $cp["List"] = $List
                $cp["Item"] = $Item
                $cp["Force"] = $Force
                $changeScript = [scriptblock]{
                    param(
                        [hashtable]$yamlHt,
                        [hashtable]$changeParameters
                    )
                    # important to return a count greater than zero, to determine if the
                    # frontmatter has changed, and therefore when updated the file must be
                    # rewritten
                    [int]$c = 0
                    if ($changeParameters["Force"]) {
                        # always count of force as 1 change
                        $yamlHt[$changeParameters["List"]] = @($changeParameters["Item"])
                        $c++
                    } else {

                        if (!($yamlHt.ContainsKey($changeParameters["List"]))) {
                            $yamlHt[$changeParameters["List"]] = @($changeParameters["Item"])
                            $c++
                        }
                        else {

                            $changeParameters["Item"] | ForEach-Object {
    
                                # need to know the type returned by powershell-yaml, so we can detect if the item
                                # is there and then to remove and count the removal
                                if ($yamlHt[$changeParameters["List"]] -is [System.Collections.IEnumerable]) {
                                    # list is an array
                                    if (!($yamlHt[$changeParameters["List"]] -contains $_)) {
                                        $yamlHt[$changeParameters["List"]] += $_
                                        $c++
                                    }
                                }
                                else {
                                    Write-Warning -Message "List is not an array, unable to add item."
                                }
                            }
    
                            
                        }

                    }
                    return $c
                }
            }
        }

        # above here we gather up the specific parameters for the Remove-MarkdownFrontmatter CmdLet, and here we call the
        # function that does the work and ensure we pass all the generic parameters that are required
        Invoke-ChangeMarkdownFrontmatter -Path $Path -UpdateDescription $updateDesc -NoRecurse:$NoRecurse -NoBackup:$NoBackup -ChangeParameters $cp -ChangeScript $changeScript 

    }
}
Set-Alias -Name amdf -Value Add-MarkdownFrontmatter
Export-ModuleMember -Function Add-MarkdownFrontmatter -Alias amdf

function Update-MarkdownFrontmatter() {
    <#
    .SYNOPSIS
        Bulk update Yaml frontmatter in Markdown files
    .DESCRIPTION
        Used to update existing properties in the Yaml frontmatter of a markdown file. If the -Force switch is used, the property will be created if it doesn't already exist.
    .NOTES
        Use -Force to add properties in the frontmatter if they don't currently exist. Without -Force the properties will only be updated if they currently exist.

        This CmdLet supports -WhatIf and -Confirm to allow you to evaluate the changes before they are made.

        If you want to completely overwrite a list property with all new values, you can call this CmdLet with -Value @("item 1","item 2","item 3") to overwrite any current list values with this new list.
    .EXAMPLE
        PS> Update-MarkdownFrontmatter -Path $VaultPath -Property "reviewstatus" -Value "Reviewed"

        Will update any markdown in the path, property "reviewstatus" with the updated value of "Reviewed". If the property doesn't exist, it won't be created.

    .EXAMPLE
        PS> Update-MarkdownFrontmatter -Path $VaultPath -Property "New List" -Value @() -Force

        Create or update the list "New List" in the frontmatter with an empty list of values
    .EXAMPLE
        PS> Update-MarkdownFrontmatter -Path $VaultPath -Property "Review Completed" -Value $true

        Update the property "Review Completed" to the value $true. If the property doesn't exist, it won't be created.
    #>
    
    
    [CmdletBinding(SupportsShouldProcess=$true)]
    param(
        # Specify the Path or List of Paths to process. Wildcards are supported.
        [Parameter(Mandatory=$true,ValueFromPipeline=$true,ValueFromPipelineByPropertyName=$true)]
        [string[]]$Path,
        # The property name to update in the frontmatter
        [Parameter(Mandatory=$true)]
        [string]$Property,
        # The value to update (capability depends on the powershell-yaml module)
        [Parameter(Mandatory=$true)]
        $Value,
        # prevent the default behaviour of recursing through subdirectories
        [switch]$NoRecurse,
        # switch to determine if backup should be skipped
        [switch]$NoBackup,
        # Force the property to be created if it doesn't already exist
        [switch]$Force
    )

    process {

        [hashtable]$cp = @{}
        [scriptblock]$changeScript = $null
        [string]$updateDesc = "Update Property"

        $cp["Property"] = $Property
        $cp["Value"] = $Value
        $cp["Force"] = $Force
        $changeScript = [scriptblock]{
            param(
                [hashtable]$yamlHt,
                [hashtable]$changeParameters
            )
            # important to return a count greater than zero, to determine if the
            # frontmatter has changed, and therefore when updated the file must be
            # rewritten
            [int]$c = 0

            if ($changeParameters["Force"]) {
                $yamlHt[$changeParameters["Property"]] = $changeParameters["Value"]
                $c++
            }
            else {
                if ($yamlHt.ContainsKey($changeParameters["Property"])) {
                    $yamlHt[$changeParameters["Property"]] = $changeParameters["Value"]
                    $c++
                }
            }

            return $c
        }

        # above here we gather up the specific parameters for the Remove-MarkdownFrontmatter CmdLet, and here we call the
        # function that does the work and ensure we pass all the generic parameters that are required
        Invoke-ChangeMarkdownFrontmatter -Path $Path -UpdateDescription $updateDesc -NoRecurse:$NoRecurse -NoBackup:$NoBackup -ChangeParameters $cp -ChangeScript $changeScript 

    }
}
Set-Alias -Name umdf -Value Update-MarkdownFrontmatter
Export-ModuleMember -Function Update-MarkdownFrontmatter -Alias umdf

function Remove-MarkdownFrontmatter() {
    <#
    .SYNOPSIS
        Remove full Properties from Markdown files, or specific items from a list in the frontmatter
    .DESCRIPTION
        Removes properties from the frontmatter of markdown files. If the property doesn't exist, it will be ignored. Can also be used to remove items from a list in the frontmatter.
    .NOTES
        This CmdLet supports -WhatIf and -Confirm to allow you to evaluate the changes before they are made.

        It is strongly recommended to use -WhatIf when first utilising this CmdLet. -NoBackup should also be avoided. You have been warned.
    .EXAMPLE
        PS> Remove-MarkdownFrontmatter -Path $VaultPath -Property "reviewstatus"

        Remove the reviewstatus property from any Markdown found in this path.
    .EXAMPLE
        PS> Remove-MarkdownFrontmatter -Path $VaultPath -List "New List" -Item "Song","Website"

        Remove the Song and Website items from the options held in the list "New List".
    .EXAMPLE
        PS> Remove-MarkdownFrontmatter -Path $VaultPath -Property "New List"

        Remove the ENTIRE property and all list items it has from all frontmatter in the path.
    #>
    
    
    [CmdletBinding(SupportsShouldProcess=$true)]
    param(
        # Specify the Path or List of Paths to process. Wildcards are supported.
        [Parameter(Mandatory=$true,ValueFromPipeline=$true,ValueFromPipelineByPropertyName=$true)]
        [string[]]$Path,
        # The properties to remove from the frontmatter Hashtable
        [Parameter(Mandatory=$true,ParameterSetName="Property")]
        [string[]]$Property,
        # When removing specific list items, specify the List and the Item
        [Parameter(Mandatory=$true,ParameterSetName="ListItem")]
        [string]$List,
        # When removing specific items from a list specify the list and the Items
        [Parameter(Mandatory=$true,ParameterSetName="ListItem")]
        [string[]]$Item,
        # prevent the default behaviour of recursing through subdirectories
        [switch]$NoRecurse,
        # switch to determine if backup should be skipped
        [switch]$NoBackup
    )
    process {

        # all this code (barring the script and parameter selection) needs to be extracted
        # from here and placed into a function that can be called from here and from
        # all the other functions that will be exposed by the module
        # this should reduce any or all boilerplate code

        [hashtable]$cp = @{}
        [scriptblock]$changeScript = $null
        [string]$updateDesc = ""

        Switch ($PSCmdLet.ParameterSetName)
        {
            "Property" {
                $updateDesc = "Remove Property"
                $cp["Property"] = $Property
                $changeScript = [scriptblock]{
                    param(
                        [hashtable]$yamlHt,
                        [hashtable]$changeParameters
                    )
                    # important to return a count greater than zero, to determine if the
                    # frontmatter has changed, and therefore when updated the file must be
                    # rewritten
                    [int]$c = 0
                    $changeParameters["Property"] | ForEach-Object {
                        if ($yamlHt.ContainsKey($_)) {
                            $yamlHt.Remove($_)
                            $c++
                        }
                    }
                    return $c
                }
            }
            "ListItem" {
                $updateDesc = "Remove List Item"
                $cp["List"] = $List
                $cp["Item"] = $Item
                $changeScript = [scriptblock]{
                    param(
                        [hashtable]$yamlHt,
                        [hashtable]$changeParameters
                    )
                    # important to return a count greater than zero, to determine if the
                    # frontmatter has changed, and therefore when updated the file must be
                    # rewritten
                    [int]$c = 0
                    $changeParameters["Item"] | ForEach-Object {
                        # do we have a list to remove from?
                        if ($yamlHt.ContainsKey($changeParameters["List"])) {

                            # need to know the type returned by powershell-yaml, so we can detect if the item
                            # is there and then to remove and count the removal
                            if ($yamlHt[$changeParameters["List"]] -is [System.Collections.IEnumerable]) {
                                # list is an array
                                if ($yamlHt[$changeParameters["List"]] -contains $_) {
                                    $yamlHt[$changeParameters["List"]] = $yamlHt[$changeParameters["List"]] -ne $_
                                    $c++
                                }
                            }
                            
                            #$yamlHt[$changeParameters["List"]] = $yamlHt[$changeParameters["List"]] -notcontains $_
                            #$c++
                        }
                    }
                    return $c
                }
            }
        }

        # above here we gather up the specific parameters for the Remove-MarkdownFrontmatter CmdLet, and here we call the
        # function that does the work and ensure we pass all the generic parameters that are required
        Invoke-ChangeMarkdownFrontmatter -Path $Path -UpdateDescription $updateDesc -NoRecurse:$NoRecurse -NoBackup:$NoBackup -ChangeParameters $cp -ChangeScript $changeScript 

    }
}
Set-Alias -Name rmdf -Value Remove-MarkdownFrontmatter
Export-ModuleMember -Function Remove-MarkdownFrontmatter -Alias rmdf

#endregion
#region "Where the real work happens - Utility functions"

function Invoke-ChangeMarkdownFrontmatter() {
    [CmdletBinding(SupportsShouldProcess=$true)]
    param(
        # Specify the Path or List of Paths to process. Wildcards are supported.
        [Parameter(Mandatory=$true)]
        [string[]]$Path,
        # The update description to be used in the WhatIf and Confirm prompts
        [Parameter(Mandatory=$true)]
        [string]$UpdateDescription,
        # prevent the default behaviour of recursing through subdirectories
        [Parameter(Mandatory=$true)]
        [switch]$NoRecurse,
        # switch to determine if backup should be skipped
        [Parameter(Mandatory=$true)]
        [switch]$NoBackup,
        # the parameters used by the scriptblock to execute any changes on the block
        [Parameter(Mandatory=$true)]
        [hashtable]$ChangeParameters,
        # the scriptblock that will be executed against the frontmatter
        [Parameter(Mandatory=$true)]
        [scriptblock]$ChangeScript
    )

    # Get the powershell-yaml module and ensure it's available
    If (!(Get-YamlModule)) {
        Write-Error -Message "Unable to get powershell-yaml module, exiting."
        return
    }

    # based on parameters, get the core frontmatter from files that _may_ be modified
    Write-Verbose "Getting Frontmatter from Path: $Path with NoRecurse: $NoRecurse"

    Write-Progress -Activity "Scanning frontmatter." -PercentComplete 0
    $frontMatter = Get-MarkdownFrontmatter -Path $Path -NoRecurse:$NoRecurse
    Write-Progress -Activity "Scanning frontmatter." -PercentComplete 100

    [int]$totalFiles = 0

    if ($null -ne $frontMatter) {
        Write-Verbose "Found $($frontMatter.Count) files with frontmatter to be checked."
        $totalFiles = $frontMatter.Count
    }
    else {
        Write-Verbose "No files with frontmatter found"
        $totalFiles = 0
        return
    }

    [int]$filesChanged = 0
    [int]$filesProcessed = 0
    [int]$totalChangeCount = 0

    Write-Progress -Activity "Changing frontmatter." -PercentComplete 0

    # check how much frontmatter is being changed
    foreach ($fm in $frontMatter) {

        $filesProcessed++
        Write-Progress -Activity "Changing frontmatter." -PercentComplete ($filesProcessed / $totalFiles * 100)

        [int]$chCount = $fm.ChangeFrontmatter($ChangeParameters, $ChangeScript)

        if ($chCount -gt 0) {
            $filesChanged++
            $totalChangeCount += $chCount
        }
    }
    Write-Progress -Activity "Changing frontmatter." -PercentComplete 100 -Completed
    Write-Verbose "Files processed: $filesProcessed, Files changed: $filesChanged, Total changes: $totalChangeCount"

    # confirm if the changes should be made
    $filesProcessed = 0
    Write-Progress -Activity "Updating files." -PercentComplete 0

    foreach($fm in $frontMatter) {

        if ($fm.IsDirty()) {

            $filesProcessed++
            Write-Progress -Activity "Updating files." -PercentComplete ($filesProcessed / $filesChanged * 100)

            if ($PSCmdLet.ShouldProcess($fm.FileInfo.FullName, $UpdateDescription)) {
                Write-Verbose "${UpdateDescription}: $($fm.FileInfo.FullName) with backup: $(!($NoBackup))"
                if (!($NoBackup)) {
                    $fm.BackupFile()
                }
                $fm.UpdateMarkdown()
            }

        }
    }

    Write-Progress -Activity "Updating files." -Completed -PercentComplete 100

}


function Get-YamlModule() {
    # based on this code:
    # https://stackoverflow.com/questions/28740320/how-do-i-check-if-a-powershell-module-is-installed/51692402#51692402

    [string]$m = "powershell-yaml"
    
    # If module is imported say that and do nothing
    if (Get-Module -Verbose:$false | Where-Object {$_.Name -eq $m}) {
        Write-Verbose "Module $m is already imported."
    }
    else {

        # If module is not imported, but available on disk then import
        if (Get-Module -ListAvailable -Verbose:$false | Where-Object {$_.Name -eq $m}) {
            Write-Verbose "Module $m is available on disk, importing."
            Import-Module $m -Verbose:$false
        }
        else {

            # If module is not imported, not available on disk, but is in online gallery then install and import
            if (Find-Module -Name $m -Verbose:$false | Where-Object {$_.Name -eq $m}) {
                Write-Verbose "Module $m is available in an online gallery, installing and importing."
                Install-Module -Name $m -Force -Scope CurrentUser -Verbose:$false
                Import-Module $m -Verbose:$false
            }
            else {

                # If the module is not imported, not available and not in the online gallery then abort
                Write-Verbose "Module $m not imported, not available and not in an online gallery, exiting."
            
            }
        }
    }

    # return true if module is now available, otherwise false
    return (Get-Module | Where-Object {$_.Name -eq $m})

}
#endregion